from collections import defaultdict
from collections.abc import Sequence
from random import shuffle
from typing import TYPE_CHECKING

import numpy as np
from pyDOE import lhs
from pygmo import fast_non_dominated_sorting as nds
from pygmo import hypervolume as hv
from pygmo import non_dominated_front_2d as nd2
from tqdm import tqdm, tqdm_notebook

from pyrvea.OtherTools.plotlyanimate import animate_init_, animate_next_
from pyrvea.OtherTools.IsNotebook import IsNotebook

if TYPE_CHECKING:
    from pyrvea.Problem.baseProblem import baseProblem
    from pyrvea.EAs.baseEA import BaseEA


class Population:
    """Define the population."""

    def __init__(
        self,
        problem: "baseProblem",
        assign_type: str = "RandomAssign",
        plotting: bool = True,
        *args
    ):
        """Initialize the population.

        Parameters
        ----------
        problem : baseProblem
            An object of the class Problem
        assign_type : str, optional
            Define the method of creation of population.
            If 'assign_type' is 'RandomDesign' the population is generated
            randomly. If 'assign_type' is 'LHSDesign', the population is
            generated via Latin Hypercube Sampling. If 'assign_type' is
            'custom', the population is imported from file. If assign_type
            is 'empty', create blank population. (the default is "RandomAssign")
        plotting : bool, optional
            (the default is True, which creates the plots)

        """
        num_var = problem.num_of_variables
        self.lower_limits = np.asarray(problem.lower_limits)
        self.upper_limits = np.asarray(problem.upper_limits)
        self.hyp = 0
        self.non_dom = 0
        self.problem = problem
        self.filename = problem.name + "_" + str(problem.num_of_objectives)
        self.plotting = plotting
        # These attributes contain the solutions.
        self.individuals = np.empty((0, num_var), float)
        self.objectives = np.empty((0, self.problem.num_of_objectives), float)
        self.fitness = np.empty((0, self.problem.num_of_objectives), float)
        self.constraint_violation = np.empty(
            (0, self.problem.num_of_constraints), float
        )
        self.individuals_archive = defaultdict(np.ndarray)
        self.objectives_arvhive = defaultdict(np.ndarray)
        self.ideal_fitness = np.full((1, self.problem.num_of_objectives), np.inf)
        self.worst_fitness = -1 * self.ideal_fitness
        self.create_new_individuals("LHSDesign")

    def create_new_individuals(
        self, design: str = "LHSDesign", pop_size: int = None, decision_variables=None
    ):
        """Create, evaluate and add new individuals to the population. Initiate Plots.

        The individuals can be created randomly, by LHS design, or can be passed by the
        user.

        Parameters
        ----------
        design : str, optional
            Describe the method of creation of new individuals.
            "RandomDesign" creates individuals randomly.
            "LHSDesign" creates individuals using Latin hypercube sampling.
        pop_size : int, optional
            Number of individuals in the population. If none, some default population
            size based on number of objectives is chosen.
        decision_variables : numpy array or list, optional
            Pass decision variables to be added to the population.
        """
        if decision_variables is not None:
            pass
        if pop_size is None:
            pop_size_options = [50, 105, 120, 126, 132, 112, 156, 90, 275]
            pop_size = pop_size_options[self.problem.num_of_objectives - 2]
        num_var = self.individuals.shape[1]
        if design == "RandomDesign":
            individuals = np.random.random((pop_size, num_var))
            # Scaling
            individuals = (
                individuals * (self.upper_limits - self.lower_limits)
                + self.lower_limits
            )
        elif design == "LHSDesign":
            individuals = lhs(num_var, samples=pop_size)
            # Scaling
            individuals = (
                individuals * (self.upper_limits - self.lower_limits)
                + self.lower_limits
            )
        else:
            print("Design not yet supported.")
        self.add(individuals)
        if self.plotting:
            self.figure = []
            self.plot_init_()

    def eval_fitness(self):
        """
        Calculate fitness based on objective values. Fitness = obj if minimized.
        """
        fitness = self.objectives * self.problem.objs
        return fitness

    def add(self, new_pop: np.ndarray):
        """Evaluate and add individuals to the population. Update ideal and nadir point.

        Parameters
        ----------
        new_pop: np.ndarray
            Decision variable values for new population.
        """
        if new_pop.ndim == 1:
            self.append_individual(new_pop)
        elif new_pop.ndim == 2:
            for ind in new_pop:
                self.append_individual(ind)
        else:
            print("Error while adding new individuals. Check dimensions.")
        # print(self.ideal_fitness)
        self.update_ideal_and_nadir()

    def keep(self, indices: list):
        """Remove individuals from population which are not in "indices".

        Parameters
        ----------
        indices: list
            Indices of individuals to keep
        """

        new_pop = self.individuals[indices, :]
        new_obj = self.objectives[indices, :]
        new_fitness = self.fitness[indices, :]
        new_CV = self.constraint_violation[indices, :]
        self.individuals = new_pop
        self.objectives = new_obj
        self.individuals_archive[len(self.individuals_archive) + 1] = new_pop
        self.objectives_arvhive[len(self.objectives_arvhive) + 1] = new_obj
        self.fitness = new_fitness
        self.constraint_violation = new_CV

    def append_individual(self, ind: np.ndarray):
        """Evaluate and add individual to the population.

        Parameters
        ----------
        ind: np.ndarray
        """
        self.individuals = np.vstack((self.individuals, ind))
        obj, CV, fitness = self.evaluate_individual(ind)
        self.objectives = np.vstack((self.objectives, obj))
        self.constraint_violation = np.vstack((self.constraint_violation, CV))
        self.fitness = np.vstack((self.fitness, fitness))

    def evaluate_individual(self, ind: np.ndarray):
        """Evaluate individual.

        Returns objective values, constraint violation, and fitness.

        Parameters
        ----------
        ind: np.ndarray
        """
        obj = self.problem.objectives(ind)
        CV = np.empty((0, self.problem.num_of_constraints), float)
        fitness = obj
        if self.problem.num_of_constraints:
            CV = self.problem.constraints(ind, obj)
            fitness = self.eval_fitness(ind, obj, self.problem)
        return (obj, CV, fitness)

    def evolve(self, EA: "BaseEA" = None, EA_parameters: dict = None) -> "Population":
        """Evolve the population with interruptions.

        Evolves the population based on the EA sent by the user.

        Parameters
        ----------
        EA: "BaseEA"
            Should be a derivative of BaseEA (Default value = None)
        EA_parameters: dict
            Contains the parameters needed by EA (Default value = None)

        """
        ##################################
        # To determine whether running in console or in notebook. Used for TQDM.
        # TQDM will be removed in future generations as number of iterations can vary
        if IsNotebook():
            progressbar = tqdm_notebook
        else:
            progressbar = tqdm
        ####################################
        # A basic evolution cycle. Will be updated to optimize() in future versions.
        ea = EA(self)
        iterations = ea.params["iterations"]
        if self.plotting:
            self.plot_objectives(1)  # Figure was created in init
        for i in progressbar(range(1, iterations), desc="Iteration"):
            ea._run_interruption(self)
            ea._next_iteration(self)
            if self.plotting:
                self.plot_objectives(i + 1)

    def mate(self):
        """Conduct crossover and mutation over the population.

        Conduct simulated binary crossover and bounded polunomial mutation.
        """
        pop = self.individuals
        pop_size, num_var = pop.shape
        shuffled_ids = list(range(pop_size))
        shuffle(shuffled_ids)
        mating_pop = pop[shuffled_ids]
        if pop_size % 2 == 1:
            # Maybe it should be pop_size-1?
            mating_pop = np.vstack((mating_pop, mating_pop[0]))
            pop_size = pop_size + 1
        # The rest closely follows the matlab code.
        ProC = 1
        ProM = 1 / num_var
        DisC = 30
        DisM = 20
        offspring = np.zeros_like(mating_pop)  # empty_like() more efficient?
        for i in range(0, pop_size, 2):
            beta = np.zeros(num_var)
            miu = np.random.rand(num_var)
            beta[miu <= 0.5] = (2 * miu[miu <= 0.5]) ** (1 / (DisC + 1))
            beta[miu > 0.5] = (2 - 2 * miu[miu > 0.5]) ** (-1 / (DisC + 1))
            beta = beta * ((-1) ** np.random.randint(0, high=2, size=num_var))
            beta[np.random.rand(num_var) > ProC] = 1  # It was in matlab code
            avg = (mating_pop[i] + mating_pop[i + 1]) / 2
            diff = (mating_pop[i] - mating_pop[i + 1]) / 2
            offspring[i] = avg + beta * diff
            offspring[i + 1] = avg - beta * diff
        min_val = np.ones_like(offspring) * self.lower_limits
        max_val = np.ones_like(offspring) * self.upper_limits
        k = np.random.random(offspring.shape)
        miu = np.random.random(offspring.shape)
        temp = np.logical_and((k <= ProM), (miu < 0.5))
        offspring_scaled = (offspring - min_val) / (max_val - min_val)
        offspring[temp] = offspring[temp] + (
            (max_val[temp] - min_val[temp])
            * (
                (
                    2 * miu[temp]
                    + (1 - 2 * miu[temp]) * (1 - offspring_scaled[temp]) ** (DisM + 1)
                )
                ** (1 / (DisM + 1))
                - 1
            )
        )
        temp = np.logical_and((k <= ProM), (miu >= 0.5))
        offspring[temp] = offspring[temp] + (
            (max_val[temp] - min_val[temp])
            * (
                1
                - (
                    2 * (1 - miu[temp])
                    + 2 * (miu[temp] - 0.5) * offspring_scaled[temp] ** (DisM + 1)
                )
                ** (1 / (DisM + 1))
            )
        )
        offspring[offspring > max_val] = max_val[offspring > max_val]
        offspring[offspring < min_val] = min_val[offspring < min_val]
        return offspring

    def plot_init_(self):
        """Initialize animation objects. Return figure"""
        obj = self.objectives
        self.figure = animate_init_(obj, self.filename + ".html")
        return self.figure

    def plot_objectives(self, iteration: int):
        """Plot the objective values of individuals in notebook. This is a hack.

        Parameters
        ----------
        iteration: int
            Iteration count.
        """
        obj = self.objectives
        self.figure = animate_next_(
            obj, self.figure, self.filename + ".html", iteration
        )

    def hypervolume(self, ref_point):
        """Calculate hypervolume. Uses package pygmo. Add checks to prevent errors.

        Parameters
        ----------
        ref_point

        Returns
        -------

        """
        non_dom = self.non_dom
        if not isinstance(ref_point, (Sequence, np.ndarray)):
            num_obj = non_dom.shape[1]
            ref_point = [ref_point] * num_obj
        non_dom = non_dom[np.all(non_dom < ref_point, axis=1), :]
        hyp = hv(non_dom)
        self.hyp = hyp.compute(ref_point)
        return self.hyp

    def non_dominated(self):
        """Fix this. check if nd2 and nds mean the same thing"""
        obj = self.objectives
        num_obj = obj.shape[1]
        if num_obj == 2:
            non_dom_front = nd2(obj)
        else:
            non_dom_front = nds(obj)
        if isinstance(non_dom_front, tuple):
            self.non_dom = self.objectives[non_dom_front[0][0]]
        elif isinstance(non_dom_front, np.ndarray):
            self.non_dom = self.objectives[non_dom_front]
        else:
            print("Non Dom error Line 285 in population.py")
        return non_dom_front

    def update_ideal_and_nadir(self, new_objective_vals: list = None):
        """Updates self.ideal and self.nadir in the fitness space.

        Uses the entire population if new_objective_vals is none.

        Parameters
        ----------
        new_objective_vals : list, optional
            Objective values for a newly added individual (the default is None, which
            calculated the ideal and nadir for the entire population.)

        """
        if new_objective_vals is None:
            check_ideal_with = self.fitness
        else:
            check_ideal_with = new_objective_vals
        self.ideal_fitness = np.amin(
            np.vstack((self.ideal_fitness, check_ideal_with)), axis=0
        )
        self.worst_fitness = np.amax(
            np.vstack((self.worst_fitness, check_ideal_with)), axis=0
        )



from itertools import combinations

import numpy as np
from scipy.special import comb
from itertools import product


def normalize(vectors):
    """
    Normalize a set of vectors.

    The length of the returned vectors will be unity.

    Parameters
    ----------
    vectors : np.ndarray
        Set of vectors of any length, except zero.

    """
    if len(vectors.shape) == 1:
        return vectors / np.linalg.norm(vectors)
    norm = np.linalg.norm(vectors, axis=1)
    return vectors / norm[:, np.newaxis]


def shear(vectors, degrees: float = 5):
    """
    Shear a set of vectors lying on the plane z=0 towards the z-axis, such that the
    resulting vectors 'degrees' angle away from the z axis.

    z is the last element of the vector, and has to be equal to zero.

    Parameters
    ----------
    vectors : numpy.ndarray
        The final element of each vector should be zero.
    degrees : float, optional
        The angle that the resultant vectors make with the z axis. Unit is radians.
        (the default is 5)

    """
    angle = degrees * np.pi / 180
    m = 1 / np.tan(angle)
    norm = np.linalg.norm(vectors, axis=1)
    vectors[:, -1] += norm * m
    return normalize(vectors)


def rotate(initial_vector, rotated_vector, other_vectors):
    """Calculate the rotation matrix that rotates the initial_vector to the
    rotated_vector. Apply that rotation on other_vectors and return.
    Uses Householder reflections twice to achieve this."""

    init_vec_norm = normalize(initial_vector)
    rot_vec_norm = normalize(rotated_vector)
    middle_vec_norm = normalize(init_vec_norm + rot_vec_norm)
    first_reflector = init_vec_norm - middle_vec_norm
    second_reflector = middle_vec_norm - rot_vec_norm
    Q1 = householder(first_reflector)
    Q2 = householder(second_reflector)
    reflection_matrix = np.matmul(Q2, Q1)
    rotated_vectors = np.matmul(other_vectors, np.transpose(reflection_matrix))
    return rotated_vectors


def householder(vector):
    """Return reflection matrix via householder transformation."""
    identity_mat = np.eye(len(vector))
    v = vector[np.newaxis]
    denominator = np.matmul(v, v.T)
    numerator = np.matmul(v.T, v)
    rot_mat = identity_mat - (2 * numerator / denominator)
    return rot_mat


def rotate_toward(initial_vector, final_vector, other_vectors, degrees: float = 5):
    """
    Rotate other_vectors (with the centre at initial_vector) towards final_vector
    by an angle degrees.

    Parameters
    ----------
    initial_vector : np.ndarray
        Centre of the vectors to be rotated.
    final_vector : np.ndarray
        The final position of the center of other_vectors.
    other_vectors : np.ndarray
        The array of vectors to be rotated
    degrees : float, optional
        The amount of rotation (the default is 5)

    Returns
    -------
    rotated_vectors : np.ndarray
        The rotated vectors
    reached: bool
        True if final_vector has been reached
    """
    final_vector = normalize(final_vector)
    initial_vector = normalize(initial_vector)
    cos_phi = np.dot(initial_vector, final_vector)
    theta = degrees * np.pi / 180
    cos_theta = np.cos(theta)
    phi = np.arccos(cos_phi)
    if phi < theta:
        return (rotate(initial_vector, final_vector, other_vectors), True)
    cos_phi_theta = np.cos(phi - theta)
    A = np.asarray([[cos_phi, 1], [1, cos_phi]])
    B = np.asarray([cos_phi_theta, cos_theta])
    x = np.linalg.solve(A, B)
    rotated_vector = x[0] * initial_vector + x[1] * final_vector
    return (rotate(initial_vector, rotated_vector, other_vectors), False)


class ReferenceVectors:
    """Class object for reference vectors."""

    def __init__(
        self,
        lattice_resolution: int,
        number_of_objectives: int,
        creation_type: str = "Uniform",
        vector_type: str = "Spherical",
        ref_point: list = None,
    ):
        """Create a Reference vectors object.

        A simplex lattice is formed

        Parameters
        ----------
        lattice_resolution : int
            Number of divisions along an axis when creating the simplex lattice.
        number_of_objectives : int
            Number of objectives.
        creation_type : str, optional
            'Uniform' creates the reference vectors uniformly using simplex lattice
            design. 'Focused' creates reference vectors symmetrically around a central
            reference vector. 'Reversed' coming soon.By default 'Uniform'.
        vector_type : str, optional
            'Spherical' normalizes the vectors to a hypersphere, i.e. the second norm
            is equal to 1. 'Planar' normalizes vectors to a plane, i.e. the first norm
            is equal to 1. By default 'Spherical'.
        ref_point : list, optional
            User preference information for a priori methods.
        """
        number_of_vectors = comb(
            lattice_resolution + number_of_objectives - 1,
            number_of_objectives - 1,
            exact=True,
        )
        self.number_of_objectives = number_of_objectives
        self.lattice_resolution = lattice_resolution
        self.number_of_vectors = number_of_vectors
        self.creation_type = creation_type
        self.vector_type = vector_type
        self.values = []
        self.ref_point = ref_point
        self._create(creation_type)
        self.values_planar = np.copy(self.values)
        self.normalize()
        self.initial_values = np.copy(self.values)
        self.initial_values_planar = np.copy(self.values_planar)
        self.neighbouring_angles()
        # self.iteractive_adapt_1() Can use this for a priori preferences!

    def _create(self, creation_type: str = "Uniform"):
        """Create the reference vectors.

        Parameters
        ----------
        creation_type : str, optional
            'Uniform' creates the reference vectors uniformly using simplex lattice
            design. 'Focused' creates reference vectors symmetrically around a central
            reference vector. By default 'Uniform'.
        """
        if creation_type == "Uniform":
            temp1 = range(1, self.number_of_objectives + self.lattice_resolution)
            temp1 = np.array(list(combinations(temp1, self.number_of_objectives - 1)))
            temp2 = np.array(
                [range(self.number_of_objectives - 1)] * self.number_of_vectors
            )
            temp = temp1 - temp2 - 1
            weight = np.zeros(
                (self.number_of_vectors, self.number_of_objectives), dtype=int
            )
            weight[:, 0] = temp[:, 0]
            for i in range(1, self.number_of_objectives - 1):
                weight[:, i] = temp[:, i] - temp[:, i - 1]
            weight[:, -1] = self.lattice_resolution - temp[:, -1]
            self.values = weight / self.lattice_resolution
            return
        elif creation_type == "Focused":
            point_set = [[0, 1, -1]] * (self.number_of_objectives - 1)
            # The cartesian product of point_set.
            initial = np.array(list(product(*point_set)))[1:]
            # First element was removed because of the error during normalization.
            initial = normalize(initial)
            initial = np.hstack((initial, np.zeros((initial.shape[0], 1))))
            final = shear(initial)
            # Adding the first element back
            final = np.vstack(([0] * (self.number_of_objectives - 1) + [1], final))
            self.values = rotate(final[0], self.ref_point, final)

    def normalize(self):
        """Normalize the reference vectors to a unit hypersphere."""
        self.number_of_vectors = self.values.shape[0]
        norm_2 = np.linalg.norm(self.values, axis=1)
        norm_1 = np.sum(self.values_planar, axis=1)
        norm_2 = np.repeat(norm_2, self.number_of_objectives).reshape(
            self.number_of_vectors, self.number_of_objectives
        )
        norm_1 = np.repeat(norm_1, self.number_of_objectives).reshape(
            self.number_of_vectors, self.number_of_objectives
        )
        self.values = np.divide(self.values, norm_1)
        self.values_planar = np.divide(self.values_planar, norm_2)

    def neighbouring_angles(self) -> np.ndarray:
        """Calculate neighbouring angles for normalization."""
        cosvv = np.dot(self.values, self.values.transpose())
        cosvv.sort(axis=1)
        cosvv = np.flip(cosvv, 1)
        acosvv = np.arccos(cosvv[:, 1])
        self.neighbouring_angles_current = acosvv
        return acosvv

    def adapt(self, fitness: np.ndarray):
        """Adapt reference vectors. Then normalize.

        Parameters
        ----------
        fitness : np.ndarray
        """
        max_val = np.amax(fitness, axis=0)
        min_val = np.amin(fitness, axis=0)
        self.values = np.multiply(
            self.initial_values,
            np.tile(np.subtract(max_val, min_val), (self.number_of_vectors, 1)),
        )
        self.normalize()

    def iteractive_adapt_1(self, ref_point, translation_param=0.2):
        """Adapt reference vectors linearly towards a reference point. Then normalize.

        The details can be found in the following paper: Hakanen, Jussi &
        Chugh, Tinkle & Sindhya, Karthik & Jin, Yaochu & Miettinen, Kaisa.
        (2016). Connections of Reference Vectors and Different Types of
        Preference Information in Interactive Multiobjective Evolutionary
        Algorithms.

        Parameters
        ----------
        ref_point :

        translation_param :
            (Default value = 0.2)

        """
        self.values = self.initial_values * translation_param + (
            (1 - translation_param) * ref_point
        )
        self.values_planar = self.initial_values_planar * translation_param + (
            (1 - translation_param) * ref_point
        )
        self.normalize()

    def slow_interactive_adapt(self, ref_point):
        """Basically a wrapper around rotate_toward. Slowly rotate ref vectors toward
        ref_point. Return a boolean value to tell if the ref_point has been reached.
        
        Parameters
        ----------
        ref_point : list or np.ndarray
            The reference vectors will slowly move towards the ref_point.
                
        Returns
        -------
        boolean
            True if ref_point has been reached. False otherwise.
        """
        assert self.creation_type == 'Focused'
        self.values, reached = rotate_toward(self.values[0], ref_point, self.values)
        self.normalize()
        return reached

    def add_edge_vectors(self):
        """Add edge vectors to the list of reference vectors.

        Used to cover the entire orthant when preference information is
        provided.

        """
        edge_vectors = np.eye(self.values.shape[1])
        self.values = np.vstack([self.values, edge_vectors])
        self.values_planar = np.vstack([self.values_planar, edge_vectors])
        self.number_of_vectors = self.values.shape[0]
        self.normalize()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyrvea.Population.Population import Population


class BaseEA:
    """This class provides the basic structure for Evolutionary algorithms."""

    def __init__(self):
        """Initialize EA here. Set up parameters, create EA specific objects."""
        pass

    def set_params(self):
        """Set up the parameters. Save in self.params"""
        pass

    def _next_gen(self):
        """Run one generation of an EA. Change nothing about the parameters."""
        pass

    def _next_iteration(self):
        """Run one iteration (a number of generations) of EA."""
        pass

    def _run_interruption(self):
        """Run interruptions in between iterations. You can update parameters/EA
        specific classes here.
        """
        pass


class BaseDecompositionEA(BaseEA):
    """This class provides the basic structure for decomposition based
    Evolutionary algorithms, such as RVEA or NSGA-III.
    """

    def __init__(self, population: "Population", EA_parameters: dict = None):
        """Initialize a Base Decomposition EA.

        This will call methods to set up the parameters of RVEA, create
        Reference Vectors, and (as of Feb 2019) run the first iteration of RVEA.

        Parameters
        ----------
        population : "Population"
            This variable is updated as evolution takes place
        EA_parameters : dict
            Takes the EA parameters

        Returns
        -------
        Population:
            Returns the Population after evolution.
        """
        self.params = self.set_params(population, EA_parameters)
        # print("Using BaseDecompositionEA init")
        self._next_iteration(population)

    def _next_iteration(self, population: "Population"):
        """Run one iteration of EA.

        One iteration consists of a constant or variable number of
        generations. This method leaves EA.params unchanged, except the current
        iteration count and gen count.

        Parameters
        ----------
        population : "Population"
            Contains current population
        """
        self.params["current_iteration_gen_count"] = 1
        while self.continue_iteration():
            self._next_gen(population)
            self.params["current_iteration_gen_count"] += 1
        self.params["current_iteration_count"] += 1

    def _next_gen(self, population: "Population"):
        """Run one generation of decomposition based EA.

        This method leaves method.params unchanged. Intended to be used by
        next_iteration.

        Parameters
        ----------
        population: "Population"
            Population object
        """
        offspring = population.mate()
        population.add(offspring)
        selected = self.select(population)
        population.keep(selected)

    def select(self, population) -> list:
        """Describe a selection mechanism. Return indices of selected
        individuals.

        Parameters
        ----------
        population : Population
            Contains the current population and problem
            information.

        Returns
        -------
        list
            List of indices of individuals to be selected.
        """
        pass

    def continue_iteration(self):
        """Checks whether the current iteration should be continued or not."""
        return self.params["current_iteration_gen_count"] <= self.params["generations"]

    def continue_evolution(self) -> bool:
        """Checks whether the current iteration should be continued or not."""
        pass



import numpy as np
from pygmo import fast_non_dominated_sorting as nds
from warnings import warn
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyrvea.allclasses import ReferenceVectors


def NSGAIII_select(
    fitness: list,
    ref_dirs: "ReferenceVectors",
    ideal_point: list = None,
    worst_point: list = None,
    extreme_points: list = None,
    n_survive: int = None,
):
    # Calculating fronts and ranks
    fronts, dl, dc, rank = nds(fitness)
    non_dominated = fronts[0]

    # Calculating worst points
    worst_of_population = np.amax(fitness, axis=0)
    worst_of_front = np.max(fitness[non_dominated, :], axis=0)
    extreme_points = get_extreme_points_c(
        fitness[non_dominated, :], ideal_point, extreme_points=extreme_points
    )
    nadir_point = get_nadir_point(
        extreme_points, ideal_point, worst_point, worst_of_population, worst_of_front
    )

    # Finding individuals in first 'n' fronts
    selection = np.asarray([], dtype=int)
    for front_id in range(len(fronts)):
        if len(np.concatenate(fronts[: front_id + 1])) < n_survive:
            continue
        else:
            fronts = fronts[: front_id + 1]
            selection = np.concatenate(fronts)
            break

    F = fitness[selection]

    last_front = fronts[-1]

    # Selecting individuals from the last acceptable front.
    if len(selection) > n_survive:
        niche_of_individuals, dist_to_niche = associate_to_niches(
            F, ref_dirs, ideal_point, nadir_point
        )
        # if there is only one front
        if len(fronts) == 1:
            n_remaining = n_survive
            until_last_front = np.array([], dtype=np.int)
            niche_count = np.zeros(len(ref_dirs), dtype=np.int)

        # if some individuals already survived
        else:
            until_last_front = np.concatenate(fronts[:-1])
            id_until_last_front = list(range(len(until_last_front)))
            niche_count = calc_niche_count(
                len(ref_dirs), niche_of_individuals[id_until_last_front]
            )
            n_remaining = n_survive - len(until_last_front)

        last_front_selection_id = list(range(len(until_last_front), len(selection)))
        if np.any(selection[last_front_selection_id] != last_front):
            print("error!!!")
        selected_from_last_front = niching(
            fitness[last_front, :],
            n_remaining,
            niche_count,
            niche_of_individuals[last_front_selection_id],
            dist_to_niche[last_front_selection_id],
        )
        final_selection = np.concatenate(
            (until_last_front, last_front[selected_from_last_front])
        )
        if extreme_points is None:
            print('Error')
        if final_selection is None:
            print('Error')
    else:
        final_selection = selection
    return(final_selection.astype(int), extreme_points)


def get_extreme_points_c(F, ideal_point, extreme_points=None):
    """Taken from pymoo"""
    # calculate the asf which is used for the extreme point decomposition
    asf = np.eye(F.shape[1])
    asf[asf == 0] = 1e6

    # add the old extreme points to never loose them for normalization
    _F = F
    if extreme_points is not None:
        _F = np.concatenate([extreme_points, _F], axis=0)

    # use __F because we substitute small values to be 0
    __F = _F - ideal_point
    __F[__F < 1e-3] = 0

    # update the extreme points for the normalization having the highest asf value each
    F_asf = np.max(__F * asf[:, None, :], axis=2)
    I = np.argmin(F_asf, axis=1)
    extreme_points = _F[I, :]
    return extreme_points


def get_nadir_point(
    extreme_points, ideal_point, worst_point, worst_of_front, worst_of_population
):
    LinAlgError = np.linalg.LinAlgError
    try:

        # find the intercepts using gaussian elimination
        M = extreme_points - ideal_point
        b = np.ones(extreme_points.shape[1])
        plane = np.linalg.solve(M, b)
        intercepts = 1 / plane

        nadir_point = ideal_point + intercepts

        if (
            not np.allclose(np.dot(M, plane), b)
            or np.any(intercepts <= 1e-6)
            or np.any(nadir_point > worst_point)
        ):
            raise LinAlgError()

    except LinAlgError:
        nadir_point = worst_of_front

    b = nadir_point - ideal_point <= 1e-6
    nadir_point[b] = worst_of_population[b]
    return nadir_point


def niching(F, n_remaining, niche_count, niche_of_individuals, dist_to_niche):
    survivors = []

    # boolean array of elements that are considered for each iteration
    mask = np.full(F.shape[0], True)

    while len(survivors) < n_remaining:

        # all niches where new individuals can be assigned to
        next_niches_list = np.unique(niche_of_individuals[mask])

        # pick a niche with minimum assigned individuals - break tie if necessary
        next_niche_count = niche_count[next_niches_list]
        next_niche = np.where(next_niche_count == next_niche_count.min())[0]
        next_niche = next_niches_list[next_niche]
        next_niche = next_niche[np.random.randint(0, len(next_niche))]

        # indices of individuals that are considered and assign to next_niche
        next_ind = np.where(np.logical_and(niche_of_individuals == next_niche, mask))[0]

        # shuffle to break random tie (equal perp. dist) or select randomly
        np.random.shuffle(next_ind)

        if niche_count[next_niche] == 0:
            next_ind = next_ind[np.argmin(dist_to_niche[next_ind])]
        else:
            # already randomized through shuffling
            next_ind = next_ind[0]

        mask[next_ind] = False
        survivors.append(int(next_ind))

        niche_count[next_niche] += 1

    return survivors


def associate_to_niches(F, ref_dirs, ideal_point, nadir_point, utopian_epsilon=0.0):
    utopian_point = ideal_point - utopian_epsilon

    denom = nadir_point - utopian_point
    denom[denom == 0] = 1e-12

    # normalize by ideal point and intercepts
    N = (F - utopian_point) / denom
    dist_matrix = calc_perpendicular_distance(N, ref_dirs)

    niche_of_individuals = np.argmin(dist_matrix, axis=1)
    dist_to_niche = dist_matrix[np.arange(F.shape[0]), niche_of_individuals]

    return niche_of_individuals, dist_to_niche


def calc_niche_count(n_niches, niche_of_individuals):
    niche_count = np.zeros(n_niches, dtype=np.int)
    index, count = np.unique(niche_of_individuals, return_counts=True)
    niche_count[index] = count
    return niche_count


def calc_perpendicular_distance(N, ref_dirs):
    u = np.tile(ref_dirs, (len(N), 1))
    v = np.repeat(N, len(ref_dirs), axis=0)

    norm_u = np.linalg.norm(u, axis=1)

    scalar_proj = np.sum(v * u, axis=1) / norm_u
    proj = scalar_proj[:, None] * u / norm_u[:, None]
    val = np.linalg.norm(proj - v, axis=1)
    matrix = np.reshape(val, (len(N), len(ref_dirs)))

    return matrix



from itertools import product

import numpy as np

from pyrvea.OtherTools.ReferenceVectors import ReferenceVectors


class newRV(ReferenceVectors):
    """pass"""

    def rotate_to_axis(self, ref_point):
        self.values = rotate(ref_point, [0, 0, 1], self.values)

    def revert_rotation(self, ref_point):
        self.values = rotate([0, 0, 1], ref_point, self.values)

    def project_to_hyperplane(self):
        """Projects the reference vectors to the hyperplane xn = 1."""
        self.values[:, -1] = 1

    def translate_to_hypersphere(self):
        """Reverse of preject_to_hyperplane()."""
        self.values[:, -1] = np.sqrt(
            1 - np.sum(np.square(self.values[:, 0:-1]), axis=1)
        )

    def interact_v2(self, ref_point):
        """New kind of interaction."""
        self.rotate_to_axis(ref_point)
        self.project_to_hyperplane()
        newvals = dist_based_translation(self.values[:, 0:-1])
        self.values[:, 0:-1] = newvals
        self.translate_to_hypersphere()
        self.revert_rotation(ref_point)

    def interact_v3(self, ref_point):
        """New kind of interaction. More coverage."""
        self.values = np.zeros(self.values.shape)
        newvals = []
        combinations = np.asarray(list(product([-1, 1], repeat=self.values.shape[1])))
        combinations[:, -1] = np.unique(abs(combinations[:, -1]), axis=0)
        for combination in combinations:
            changedvals = np.copy(self.initial_values) * combination
            if newvals == []:
                newvals = changedvals
            else:
                newvals = np.vstack((newvals, changedvals))
        newvals = dist_based_translation(newvals[:, 0:-1])
        changedvals = np.ones((newvals.shape[0], newvals.shape[1] + 1))
        changedvals[:, 0:-1] = newvals
        self.values = changedvals
        self.translate_to_hypersphere()
        self.revert_rotation(ref_point)
        num_rows = self.values.shape[0]
        delete_rows = []
        for index in range(num_rows):
            if np.any(self.values[index, :] < 0):
                if delete_rows == []:
                    delete_rows = np.asarray(index)
                else:
                    delete_rows = np.hstack((delete_rows, index))
        self.values = np.delete(self.values, delete_rows, axis=0)


def rotate(initial_vector, rotated_vector, other_vectors):
    """Calculate the rotation matrix that rotates the initial_vector to the
    rotated_vector. Apply that rotation on other_vectors and return.
    Uses Householder reflections twice to achieve this."""

    init_vec_norm = normalize(initial_vector)
    rot_vec_norm = normalize(rotated_vector)
    middle_vec_norm = normalize(init_vec_norm + rot_vec_norm)
    first_reflector = init_vec_norm - middle_vec_norm
    second_reflector = middle_vec_norm - rot_vec_norm
    Q1 = householder(first_reflector)
    Q2 = householder(second_reflector)
    reflection_matrix = np.matmul(Q2, Q1)
    rotated_vectors = np.matmul(other_vectors, np.transpose(reflection_matrix))
    return rotated_vectors


def normalize(vector):
    """Normalize and return a vector."""
    return vector / np.linalg.norm(vector)


def householder(vector):
    """Return reflection matrix via householder transformation."""
    identity_mat = np.eye(len(vector))
    v = vector[np.newaxis]
    denominator = np.matmul(v, v.T)
    numerator = np.matmul(v.T, v)
    rot_mat = identity_mat - (2 * numerator / denominator)
    return rot_mat


def dist_based_translation(vectors):
    """Translates points towards origin based on distance."""
    dist = np.sqrt(np.sum(np.square(vectors), axis=1))
    max_dist = np.amax(dist)
    #max_dist = 1
    alpha = 5
    ratio = alpha * (1 / (dist * dist) - 1 / (max_dist * max_dist))
    t_factor = 1 / (1 + ratio)
    return vectors * t_factor[np.newaxis].T


def main():
    rv = newRV(20, 3)
    rv.interact_v3([1, 1, 1])


if __name__ == '__main':
    main()



class baseProblem:
    """Base class for the problems."""

    def __init__(
        self,
        name=None,
        num_of_variables=None,
        num_of_objectives=None,
        num_of_constraints=0,
        upper_limits=1,
        lower_limits=0,
    ):
        """
        Pydocstring is ruthless.

        Parameters
        ----------
        name

        num_of_variables

        num_of_objectives

        num_of_constraints

        upper_limits

        lower_limits

        """
        self.name = name
        self.num_of_variables = num_of_variables
        self.num_of_objectives = num_of_objectives
        self.num_of_constraints = num_of_constraints
        self.obj_func = []
        self.upper_limits = upper_limits
        self.lower_limits = lower_limits

    def objectives(self, decision_variables):
        """Accept a sample. Return Objective values.

        Parameters
        ----------
        decision_variables
        """
        pass

    def constraints(self, decision_variables):
        """Accept a sample and/or corresponding objective values.

        Parameters
        ----------
        decision_variables
        """
        pass

    def update():
        """Update the problem based on new information."""
        pass



from typing import TYPE_CHECKING

from pyrvea.Selection.APD_select import APD_select
from pyrvea.EAs.baseEA import BaseDecompositionEA
from pyrvea.OtherTools.ReferenceVectors import ReferenceVectors

import numpy as np

if TYPE_CHECKING:
    from pyrvea.Population.Population import Population


class RVEA(BaseDecompositionEA):
    """The python version reference vector guided evolutionary algorithm.

    See the details of RVEA in the following paper

    R. Cheng, Y. Jin, M. Olhofer and B. Sendhoff, A Reference Vector Guided
    Evolutionary Algorithm for Many-objective Optimization, IEEE Transactions on
    Evolutionary Computation, 2016

    The source code of pyrvea is implemented by Bhupinder Saini

    If you have any questions about the code, please contact:

    Bhupinder Saini: bhupinder.s.saini@jyu.fi

    Project researcher at University of Jyväskylä.
    """

    def set_params(
        self,
        population: "Population",
        population_size: int = None,
        lattice_resolution: int = None,
        interact: bool = False,
        a_priori_preference: bool = False,
        generations_per_iteration: int = 100,
        iterations: int = 10,
        Alpha: float = 2,
        plotting: bool = True,
    ):
        """Set up the parameters. Save in RVEA.params. Note, this should be
        changed to align with the current structure.

        Parameters
        ----------
        population : Population
            Population object
        population_size : int
            Population Size
        lattice_resolution : int
            Lattice resolution
        interact : bool
            bool to enable or disable interaction. Enabled if True
        a_priori_preference : bool
            similar to interact
        generations_per_iteration : int
            Number of generations per iteration.
        iterations : int
            Total Number of iterations.
        Alpha : float
            The alpha parameter of APD selection.
        plotting : bool
            Useless really.
        Returns
        -------

        """
        lattice_resolution_options = {
            "2": 49,
            "3": 13,
            "4": 7,
            "5": 5,
            "6": 4,
            "7": 3,
            "8": 3,
            "9": 3,
            "10": 3,
        }
        if population.problem.num_of_objectives < 11:
            lattice_resolution = lattice_resolution_options[
                str(population.problem.num_of_objectives)
            ]
        else:
            lattice_resolution = 3
        rveaparams = {
            "population_size": population_size,
            "lattice_resolution": lattice_resolution,
            "interact": interact,
            "a_priori": a_priori_preference,
            "generations": generations_per_iteration,
            "iterations": iterations,
            "Alpha": Alpha,
            "ploton": plotting,
            "current_iteration_gen_count": 0,
            "current_iteration_count": 0,
            "reference_vectors": ReferenceVectors(
                lattice_resolution, population.problem.num_of_objectives
            ),
        }
        return rveaparams

    def _run_interruption(self, population: "Population"):
        """Run the interruption phase of RVEA.

        Use this phase to make changes to RVEA.params or other objects.
        Updates Reference Vectors, conducts interaction with the user.

        Parameters
        ----------
        population : Population
        """
        if self.params["interact"] or (
            self.params["a_priori"] and self.params["current_iteration_count"] == 1
        ):
            ideal = population.ideal_fitness
            refpoint = np.zeros_like(ideal)
            print("Ideal vector is ", ideal)
            for index in range(len(refpoint)):
                while True:
                    print("Preference for objective ", index + 1)
                    print("Ideal value = ", ideal[index])
                    pref_val = float(
                        input("Please input a value worse than the ideal: ")
                    )
                    if pref_val > ideal[index]:
                        refpoint[index] = pref_val
                        break
            refpoint = refpoint - ideal
            norm = np.sqrt(np.sum(np.square(refpoint)))
            refpoint = refpoint / norm
            self.params["reference_vectors"].iteractive_adapt_1(refpoint)
            self.params["reference_vectors"].add_edge_vectors()
        else:
            self.params["reference_vectors"].adapt(population.fitness)
        self.params["reference_vectors"].neighbouring_angles()

    def select(self, population: "Population"):
        """Describe a selection mechanism. Return indices of selected
        individuals.

        # APD Based selection. # This is different from the paper. #
        params.genetations != total number of generations. This is a compromise.
        Also this APD uses an archived ideal point, rather than current, potentially
        worse ideal point.

        Parameters
        ----------
        population : Population
            Population information

        Returns
        -------
        list
            list: Indices of selected individuals.
        """
        penalty_factor = (
            (self.params["current_iteration_gen_count"] / self.params["generations"])
            ** self.params["Alpha"]
        ) * population.problem.num_of_objectives
        return APD_select(
            fitness=population.fitness,
            vectors=self.params["reference_vectors"],
            penalty_factor=penalty_factor,
            ideal=population.ideal_fitness,
        )



import numpy as np
from warnings import warn
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyrvea.allclasses import ReferenceVectors


def APD_select(
    fitness: list,
    vectors: "ReferenceVectors",
    penalty_factor: float,
    ideal: list = None,
):
    """Select individuals for mating on basis of Angle penalized distance.

    Args:
        fitness (list): Fitness of the current population.

        vectors (ReferenceVectors): Class containing reference vectors.

        penalty_factor (float): Multiplier of angular deviation from Reference
            vectors. See RVEA paper for details.

        ideal (list): ideal point for the population.
            Uses the min fitness value if None.

    Returns:
        [type]: A list of indices of the selected individuals.
    """
    refV = vectors.neighbouring_angles_current
    # Normalization - There may be problems here
    if ideal is not None:
        fmin = ideal
    else:
        fmin = np.amin(fitness, axis=0)
    translated_fitness = fitness - fmin
    fitness_norm = np.linalg.norm(translated_fitness, axis=1)
    fitness_norm = np.repeat(fitness_norm, len(translated_fitness[0, :])).reshape(
        len(fitness), len(fitness[0, :])
    )
    normalized_fitness = np.divide(translated_fitness, fitness_norm)  # Checked, works.
    cosine = np.dot(normalized_fitness, np.transpose(vectors.values))
    if cosine[np.where(cosine > 1)].size:
        warn("RVEA.py line 60 cosine larger than 1 decreased to 1")
        cosine[np.where(cosine > 1)] = 1
    if cosine[np.where(cosine < 0)].size:
        warn("RVEA.py line 64 cosine smaller than 0 increased to 0")
        cosine[np.where(cosine < 0)] = 0
    # Calculation of angles between reference vectors and solutions
    theta = np.arccos(cosine)
    # Reference vector asub_population_indexssignment
    assigned_vectors = np.argmax(cosine, axis=1)
    selection = np.array([], dtype=int)
    # Selection
    for i in range(0, len(vectors.values)):
        sub_population_index = np.atleast_1d(
            np.squeeze(np.where(assigned_vectors == i))
        )
        sub_population_fitness = translated_fitness[sub_population_index]
        if len(sub_population_fitness > 0):
            # APD Calculation
            angles = theta[sub_population_index, i]
            angles = np.divide(angles, refV[i])  # This is correct.
            # You have done this calculation before. Check with fitness_norm
            # Remove this horrible line
            sub_pop_fitness_magnitude = np.sqrt(
                np.sum(np.power(sub_population_fitness, 2), axis=1)
            )
            apd = np.multiply(
                np.transpose(sub_pop_fitness_magnitude),
                (1 + np.dot(penalty_factor, angles)),
            )
            minidx = np.where(apd == np.amin(apd))
            selx = sub_population_index[minidx]
            if selection.shape[0] == 0:
                selection = np.hstack((selection, np.transpose(selx[0])))
            else:
                selection = np.vstack((selection, np.transpose(selx[0])))
    return selection.squeeze()



from plotly.offline.offline import plot
import plotly.graph_objs as go
import numpy as np
import pandas as pd
from typing import Union


def animate_init_(data: Union[np.ndarray, pd.DataFrame, list], filename: str) -> dict:
    """Plot the first (or zeroth) iteration of a population.

    Intended as a frames object. Plots Scatter for 2D and 3D data.
    Plots parallel coordinate plot for higher dimensional data.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        Contains the data to be plotted. Each row is an individual's objective values.
    filename : str
        Contains the name of the file to which the plot is saved.

    Returns
    -------
    dict
        Plotly figure object
    """
    numobj = data.shape[1]
    if numobj == 2:
        figure = animate_2d_init_(data, filename)
    elif numobj == 3:
        figure = animate_3d_init_(data, filename)
    elif numobj >= 4:
        figure = animate_parallel_coords_init_(data, filename)
    print("Plot saved as: ", filename)
    print("View the plot by opening the file in browser.")
    print("To view the plot in Jupyter Notebook, use the IFrame command.")
    return figure


def animate_next_(
    data: Union[np.ndarray, pd.DataFrame, list],
    figure: dict,
    filename: str,
    generation: int,
) -> dict:
    """Plot the next set of individuals in an animation.

    Plots scatter for 2D and 3D data, parallel coordinate plot for 4D and up.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        The objective values to be plotted
    figure : dict
        Plotly figure object compatible dict
    filename : str
        Name of the file to which the plot is saved
    generation : int
        Iteration Number

    Returns
    -------
    dict
        Plotly Figure Object
    """

    numobj = data.shape[1]
    if numobj == 2:
        figure = animate_2d_next_(data, figure, filename, generation)
    elif numobj == 3:
        figure = animate_3d_next_(data, figure, filename, generation)
    elif numobj >= 4:
        figure = animate_parallel_coords_next_(data, figure, filename, generation)
    return figure


def animate_2d_init_(
    data: Union[np.ndarray, pd.DataFrame, list], filename: str
) -> dict:
    """Initiate a 2D scatter animation.
    
    Only for 2D data.
    
    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        Objective values
    filename : str
        Name of the file to which plot is saved
    
    Returns
    -------
    dict
        Plotly Figure Object
    """

    figure = {"data": [], "layout": {}, "frames": []}
    figure["layout"]["xaxis"] = {"autorange": True}
    figure["layout"]["yaxis"] = {"autorange": True}
    figure["layout"]["hovermode"] = "closest"
    figure["layout"]["sliders"] = {
        "args": ["transition", {"duration": 400, "easing": "cubic-in-out"}],
        "initialValue": "1952",
        "plotlycommand": "animate",
        "visible": True,
    }
    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "Iteration:",
            "visible": True,
            "xanchor": "right",
        },
        "transition": {"duration": 300, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": [],
    }
    figure["layout"]["sliders"] = [sliders_dict]
    data_dict = {
        "x": list(data[:, 0]),
        "y": list(data[:, 1]),
        "mode": "markers",
        "marker": {
            "size": 5,
            "color": "rgba(255, 182, 193, .9)",
            "line": dict(width=2),
        },
    }
    figure["data"].append(data_dict)
    plot(figure, filename=filename)
    animate_2d_next_(data, figure, filename, 0)
    return figure


def animate_2d_next_(
    data: Union[np.ndarray, pd.DataFrame, list],
    figure: dict,
    filename: str,
    generation: int,
) -> dict:
    """Plot the next set of individuals in a 2D scatter animation.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        The objective values to be plotted
    figure : dict
        Plotly figure object compatible dict
    filename : str
        Name of the file to which the plot is saved
    generation : int
        Iteration Number

    Returns
    -------
    dict
        Plotly Figure Object
    """
    frame = {"data": [], "name": str(generation)}
    sliders_dict = figure["layout"]["sliders"][0]
    data_dict = {
        "x": list(data[:, 0]),
        "y": list(data[:, 1]),
        "mode": "markers",
        "marker": {
            "size": 5,
            "color": "rgba(255, 182, 193, .9)",
            "line": dict(width=2),
        },
    }
    frame["data"].append(data_dict)
    figure["frames"].append(frame)
    slider_step = {
        "args": [
            [generation],
            {
                "frame": {"duration": 300, "redraw": False},
                "mode": "immediate",
                "transition": {"duration": 300},
            },
        ],
        "label": generation,
        "method": "animate",
    }
    sliders_dict["steps"].append(slider_step)
    figure["layout"]["sliders"] = [sliders_dict]
    plot(figure, auto_open=False, filename=filename)
    return figure


def animate_3d_init_(
    data: Union[np.ndarray, pd.DataFrame, list], filename: str
) -> dict:
    """Plot the first (or zeroth) iteration of a population.

    Intended as a frames object. Plots Scatter 3D data.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        Contains the data to be plotted. Each row is an individual's objective values.
    filename : str
        Contains the name of the file to which the plot is saved.

    Returns
    -------
    dict
        Plotly figure object
    """
    figure = {"data": [], "layout": {}, "frames": []}
    figure["layout"]["hovermode"] = "closest"
    figure["layout"]["sliders"] = {
        "args": ["transition", {"duration": 400, "easing": "cubic-in-out"}],
        "initialValue": "1",
        "plotlycommand": "animate",
        "visible": True,
    }
    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "Iteration:",
            "visible": True,
            "xanchor": "right",
        },
        "transition": {"duration": 300, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": [],
    }
    figure["layout"]["sliders"] = [sliders_dict]
    figure["layout"]["scene"] = dict(
        xaxis=dict(
            backgroundcolor="rgb(200, 200, 230)",
            gridcolor="rgb(255, 255, 255)",
            showbackground=True,
            zerolinecolor="rgb(255, 255, 255)",
        ),
        yaxis=dict(
            backgroundcolor="rgb(230, 200,230)",
            gridcolor="rgb(255, 255, 255)",
            showbackground=True,
            zerolinecolor="rgb(255, 255, 255)",
        ),
        zaxis=dict(
            backgroundcolor="rgb(230, 230,200)",
            gridcolor="rgb(255, 255, 255)",
            showbackground=True,
            zerolinecolor="rgb(255, 255, 255)",
        ),
    )

    data_dict = go.Scatter3d(
        x=list(data[:, 0]),
        y=list(data[:, 1]),
        z=list(data[:, 2]),
        mode="markers",
        marker=dict(
            size=8,
            color=data[:, 2],
            colorscale="Viridis",
            opacity=0.5,
            line=dict(width=2, color="black", colorscale="Viridis"),
        ),
    )
    figure["data"].append(data_dict)
    plot(figure, filename=filename)
    animate_3d_next_(data, figure, filename, 0)
    return figure


def animate_3d_next_(
    data: Union[np.ndarray, pd.DataFrame, list],
    figure: dict,
    filename: str,
    generation: int,
) -> dict:
    """Plot the next set of individuals in an animation.

    Plots scatter for 3D data.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        The objective values to be plotted
    figure : dict
        Plotly figure object compatible dict
    filename : str
        Name of the file to which the plot is saved
    generation : int
        Iteration Number

    Returns
    -------
    dict
        Plotly Figure Object
    """
    frame = {"data": [], "name": str(generation)}
    sliders_dict = figure["layout"]["sliders"][0]
    data_dict = go.Scatter3d(
        x=list(data[:, 0]),
        y=list(data[:, 1]),
        z=list(data[:, 2]),
        mode="markers",
        marker=dict(
            size=8,
            color=data[:, 2],
            colorscale="Viridis",
            opacity=0.5,
            line=dict(width=2, color="black", colorscale="Viridis"),
        ),
    )
    frame["data"].append(data_dict)
    figure["frames"].append(frame)
    slider_step = {
        "args": [
            [generation],
            {
                "frame": {"duration": 300, "redraw": False},
                "mode": "immediate",
                "transition": {"duration": 300},
            },
        ],
        "label": generation,
        "method": "animate",
    }
    sliders_dict["steps"].append(slider_step)
    figure["layout"]["sliders"] = [sliders_dict]
    plot(figure, auto_open=False, filename=filename)
    return figure


def animate_parallel_coords_init_(
    data: Union[np.ndarray, pd.DataFrame, list], filename: str
) -> dict:
    """Plot the first (or zeroth) iteration of a population.

    Intended as a frames object. Plots parallel coordinate plot for >3D data.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        Contains the data to be plotted. Each row is an individual's objective values.
    filename : str
        Contains the name of the file to which the plot is saved.

    Returns
    -------
    dict
        Plotly figure object
    """
    figure = {"data": [], "layout": {}, "frames": []}
    objectives = pd.DataFrame(data)
    figure["layout"]["hovermode"] = "closest"
    figure["layout"]["sliders"] = {
        "args": ["transition", {"duration": 400, "easing": "cubic-in-out"}],
        "initialValue": "1952",
        "plotlycommand": "animate",
        "visible": True,
    }
    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "Iteration:",
            "visible": True,
            "xanchor": "right",
        },
        "transition": {"duration": 300, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": [],
    }
    figure["layout"]["sliders"] = [sliders_dict]
    dimensions = [
        dict(
            range=[min(objectives[column]), max(objectives[column])],
            label="f" + str(column),
            values=objectives[column],
        )
        for column in objectives
    ]
    line = dict(
        color=objectives[0],
        colorscale="Viridis",
        showscale=True,
        cmin=min(objectives[objectives.columns[0]]),
        cmax=max(objectives[objectives.columns[0]]),
    )
    data_dict = go.Parcoords(line=line, dimensions=dimensions)
    figure["data"].append(data_dict)
    plot(figure, filename=filename)
    animate_parallel_coords_next_(data, figure, filename, 0)
    return figure


def animate_parallel_coords_next_(
    data: Union[np.ndarray, pd.DataFrame, list],
    figure: dict,
    filename: str,
    generation: int,
) -> dict:
    """Plot the next set of individuals in an animation.

    Plots parallel coordinate plot for 4D and up.

    Parameters
    ----------
    data : Union[np.ndarray, pd.DataFrame, list]
        The objective values to be plotted
    figure : dict
        Plotly figure object compatible dict
    filename : str
        Name of the file to which the plot is saved
    generation : int
        Iteration Number

    Returns
    -------
    dict
        Plotly Figure Object
    """
    frame = {"data": [], "name": str(generation)}
    objectives = pd.DataFrame(data)
    sliders_dict = figure["layout"]["sliders"][0]
    dimensions = [
        dict(
            range=[min(objectives[column]), max(objectives[column])],
            label="f" + str(column),
            values=objectives[column],
        )
        for column in objectives
    ]
    line = dict(
        color=objectives[0],
        colorscale="Viridis",
        showscale=True,
        cmin=min(objectives[objectives.columns[0]]),
        cmax=max(objectives[objectives.columns[0]]),
    )
    data_dict = go.Parcoords(line=line, dimensions=dimensions)
    frame["data"].append(data_dict)
    figure["frames"].append(frame)
    slider_step = {
        "args": [
            [generation],
            {
                "frame": {"duration": 300, "redraw": False},
                "mode": "immediate",
                "transition": {"duration": 300},
            },
        ],
        "label": generation,
        "method": "animate",
    }
    sliders_dict["steps"].append(slider_step)
    figure["layout"]["sliders"] = [sliders_dict]
    plot(figure, auto_open=False, filename=filename)
    return figure


def test():
    data1 = np.random.rand(100, 3)
    data2 = np.square(data1)
    data3 = np.square(data2)
    data4 = np.square(data3)
    data = [data2, data3, data4]
    filename = "firsttest.html"
    figure = animate_3d_init_(data1, filename)
    for i in range(1, 4):
        figure = animate_3d_next_(data[i - 1], figure, filename, i)


def test2():
    data = np.random.rand(100, 1) * np.pi
    obj = np.hstack((data, np.sin(data), np.cos(data), np.log(data)))
    figure = animate_parallel_coords_init_(obj, "partest.html")
    for i in range(1, 10):
        obj = np.hstack((data, np.sin(data / i), np.cos(data / i), np.log(data / i)))
        figure = animate_parallel_coords_next_(obj, figure, "partest.html", i)



from optproblems import dtlz, zdt
from pyrvea.Problem.baseProblem import baseProblem


class testProblem(baseProblem):
    """Defines the problem."""

    def __init__(
        self,
        name=None,
        num_of_variables=None,
        num_of_objectives=None,
        num_of_constraints=0,
        upper_limits=1,
        lower_limits=0,
    ):
        """Pydocstring is ruthless.

        Args:
            name:
            num_of_variables:
            num_of_objectives:
            num_of_constraints:
            upper_limits:
            lower_limits:
        """
        super(testProblem, self).__init__(
            name,
            num_of_variables,
            num_of_objectives,
            num_of_constraints,
            upper_limits,
            lower_limits,
        )
        if name == "ZDT1":
            self.obj_func = zdt.ZDT1()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "ZDT2":
            self.obj_func = zdt.ZDT2()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "ZDT3":
            self.obj_func = zdt.ZDT3()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "ZDT4":
            self.obj_func = zdt.ZDT4()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "ZDT5":
            self.obj_func = zdt.ZDT5()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "ZDT6":
            self.obj_func = zdt.ZDT6()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ1":
            self.obj_func = dtlz.DTLZ1(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ2":
            self.obj_func = dtlz.DTLZ2(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ3":
            self.obj_func = dtlz.DTLZ3(num_of_objectives, num_of_variables)
            self.lower_limits = 0
            self.upper_limits = 1
        elif name == "DTLZ4":
            self.obj_func = dtlz.DTLZ4(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ5":
            self.obj_func = dtlz.DTLZ5(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ6":
            self.obj_func = dtlz.DTLZ6(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ7":
            self.obj_func = dtlz.DTLZ7(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds

    def objectives(self, decision_variables) -> list:
        """Use this method to calculate objective functions.

        Args:
            decision_variables:
        """
        return self.obj_func(decision_variables)

    def constraints(self, decision_variables, objective_variables):
        """Calculate constraint violation.

        Args:
            decision_variables:
            objective_variables:
        """
        print("Error: Constraints not supported yet.")



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pyrvea',
 'pyrvea.EAs',
 'pyrvea.OtherTools',
 'pyrvea.Population',
 'pyrvea.Problem',
 'pyrvea.Recombination',
 'pyrvea.Selection']

package_data = \
{'': ['*']}

install_requires = \
['diversipy>=0.8.0,<0.9.0',
 'numpy>=1.16,<2.0',
 'optproblems>=1.2,<2.0',
 'pandas>=0.24.2,<0.25.0',
 'plotly>=3.8,<4.0',
 'pyDOE>=0.3.8,<0.4.0',
 'pygmo>=2.10,<3.0',
 'scipy>=1.2,<2.0',
 'tqdm>=4.31,<5.0']

setup_kwargs = {
    'name': 'pyrvea',
    'version': '0.1.1',
    'description': 'The python version reference vector guided evolutionary algorithm.',
    'long_description': '# README\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/EchoBetaGlass/pyrvea/master)\n\nThe python version reference vector guided evolutionary algorithm.\n\nCurrently supported: Multi-objective minimization with visualization and interaction support. Preference is accepted as a reference point.\n\nTo test the code, open the [binder link](https://mybinder.org/v2/gh/EchoBetaGlass/pyrvea/master) and read example.ipynb.\n\nRead the documentation [here](https://pyrvea.readthedocs.io/en/latest/)\n\n### Requirements:\n* Python 3.6 or up\n* [Poetry dependency manager](https://github.com/sdispater/poetry): Only for developers\n\n### Installation process for normal users:\n* Run: `pip install pyrvea`\n\n### Installation process for developers:\n* Download and extract the code\n* Create a new virtual environment for the project\n* If you want to take part in the development process, run `poetry install` instead.\n\n## See the details of RVEA in the following paper\n\nR. Cheng, Y. Jin, M. Olhofer and B. Sendhoff,\nA Reference Vector Guided Evolutionary Algorithm for Many-objective\nOptimization, IEEE Transactions on Evolutionary Computation, 2016\n\nThe source code of pyrvea is implemented by Bhupinder Saini\n\nIf you have any questions about the code, please contact:\n\nBhupinder Saini: bhupinder.s.saini@jyu.fi\\\nProject researcher at University of Jyväskylä.',
    'author': 'Bhupinder Saini',
    'author_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import TYPE_CHECKING

from pyrvea.Selection.NSGAIII_select import NSGAIII_select
from pyrvea.EAs.baseEA import BaseDecompositionEA
from pyrvea.OtherTools.ReferenceVectors import ReferenceVectors

import numpy as np

if TYPE_CHECKING:
    from pyrvea.Population.Population import Population


class NSGAIII(BaseDecompositionEA):
    """Python Implementation of NSGA-III. Based on the pymoo package.

    [description]
    """

    def set_params(
        self,
        population: "Population" ,
        population_size: int = None,
        lattice_resolution: int = None,
        interact: bool = True,
        a_priori_preference: bool = False,
        generations_per_iteration: int = 100,
        iterations: int = 10,
        plotting: bool = True,
    ):
        lattice_resolution_options = {
            "2": 49,
            "3": 13,
            "4": 7,
            "5": 5,
            "6": 4,
            "7": 3,
            "8": 3,
            "9": 3,
            "10": 3,
        }
        if population.problem.num_of_objectives < 11:
            lattice_resolution = lattice_resolution_options[
                str(population.problem.num_of_objectives)
            ]
        else:
            lattice_resolution = 3
        reference_vectors = ReferenceVectors(
                lattice_resolution, population.problem.num_of_objectives
            )
        nsga3params = {
            "population_size": reference_vectors.number_of_vectors,
            "lattice_resolution": lattice_resolution,
            "interact": interact,
            "a_priori": a_priori_preference,
            "generations": generations_per_iteration,
            "iterations": iterations,
            "ploton": plotting,
            "current_iteration_gen_count": 0,
            "current_iteration_count": 0,
            "reference_vectors": reference_vectors,
            "extreme_points": None,
        }
        return nsga3params

    def select(self, population: "Population"):
        Selection, extreme_points = NSGAIII_select(
            population.fitness,
            self.params["reference_vectors"].values_planar,
            population.ideal_fitness,
            population.worst_fitness,
            self.params["extreme_points"],
            self.params["population_size"]
        )
        self.params["extreme_points"] = extreme_points
        return Selection

    def _run_interruption(self, population: "Population"):
        """Run the interruption phase of RVEA.

        Use this phase to make changes to RVEA.params or other objects.
        Updates Reference Vectors, conducts interaction with the user.

        Parameters
        ----------
        population : Population
        """
        if self.params["interact"] or (
            self.params["a_priori"] and self.params["current_iteration_count"] == 1
        ):
            ideal = population.ideal_fitness
            refpoint = np.zeros_like(ideal)
            print("Ideal vector is ", ideal)
            for index in range(len(refpoint)):
                while True:
                    print("Preference for objective ", index + 1)
                    print("Ideal value = ", ideal[index])
                    pref_val = float(
                        input("Please input a value worse than the ideal: ")
                    )
                    if pref_val > ideal[index]:
                        refpoint[index] = pref_val
                        break
            refpoint = refpoint - ideal
            norm = np.sqrt(np.sum(np.square(refpoint)))
            refpoint = refpoint / norm
            self.params["reference_vectors"].iteractive_adapt_1(refpoint)
            self.params["reference_vectors"].add_edge_vectors()
        else:
            self.params["reference_vectors"].adapt(population.fitness)
        self.params["reference_vectors"].neighbouring_angles()




from plotlyanimate import animate_init_, animate_next_
import numpy as np
from itertools import product


def shear(vectors, degrees: float = 5):
    """
    Shear a set of vectors lying on the plane z=0 towards the z-axis, such that the
    resulting vectors 'degrees' angle away from the z axis.

    z is the last element of the vector, and has to be equal to zero.

    Parameters
    ----------
    vectors : numpy.ndarray
        The final element of each vector should be zero.
    degrees : float, optional
        The angle that the resultant vectors make with the z axis. Unit is radians.
        (the default is 5)

    """
    angle = degrees * np.pi / 180
    m = 1 / np.tan(angle)
    norm = np.linalg.norm(vectors, axis=1)
    vectors[:, -1] += norm * m
    return normalize(vectors)


def normalize(vectors):
    """
    Normalize a set of vectors.

    The length of the returned vectors will be unity.

    Parameters
    ----------
    vectors : np.ndarray
        Set of vectors of any length, except zero.

    """
    if len(vectors.shape) == 1:
        return vectors / np.linalg.norm(vectors)
    norm = np.linalg.norm(vectors, axis=1)
    return vectors / norm[:, np.newaxis]


def rotate(initial_vector, rotated_vector, other_vectors):
    """Calculate the rotation matrix that rotates the initial_vector to the
    rotated_vector. Apply that rotation on other_vectors and return.
    Uses Householder reflections twice to achieve this."""

    init_vec_norm = normalize(initial_vector)
    rot_vec_norm = normalize(rotated_vector)
    middle_vec_norm = normalize(init_vec_norm + rot_vec_norm)
    first_reflector = init_vec_norm - middle_vec_norm
    second_reflector = middle_vec_norm - rot_vec_norm
    Q1 = householder(first_reflector)
    Q2 = householder(second_reflector)
    reflection_matrix = np.matmul(Q2, Q1)
    rotated_vectors = np.matmul(other_vectors, np.transpose(reflection_matrix))
    return rotated_vectors


def householder(vector):
    """Return reflection matrix via householder transformation."""
    identity_mat = np.eye(len(vector))
    v = vector[np.newaxis]
    denominator = np.matmul(v, v.T)
    numerator = np.matmul(v.T, v)
    rot_mat = identity_mat - (2 * numerator / denominator)
    return rot_mat


def rotate_toward(initial_vector, final_vector, other_vectors, degrees: float = 5):
    """
    Rotate other_vectors (with the centre at initial_vector) towards final_vector
    by an angle degrees.

    Parameters
    ----------
    initial_vector : np.ndarray
        Centre of the vectors to be rotated.
    final_vector : np.ndarray
        The final position of the center of other_vectors.
    other_vectors : np.ndarray
        The array of vectors to be rotated
    degrees : float, optional
        The amount of rotation (the default is 5)

    Returns
    -------
    rotated_vectors : np.ndarray
        The rotated vectors
    reached: bool
        True if final_vector has been reached
    """
    final_vector = normalize(final_vector)
    initial_vector = normalize(initial_vector)
    cos_phi = np.dot(initial_vector, final_vector)
    theta = degrees * np.pi / 180
    cos_theta = np.cos(theta)
    phi = np.arccos(cos_phi)
    if phi < theta:
        return (rotate(initial_vector, final_vector, other_vectors), True)
    cos_phi_theta = np.cos(phi - theta)
    A = np.asarray([[cos_phi, 1], [1, cos_phi]])
    B = np.asarray([cos_phi_theta, cos_theta])
    x = np.linalg.solve(A, B)
    rotated_vector = x[0] * initial_vector + x[1] * final_vector
    return (rotate(initial_vector, rotated_vector, other_vectors), False)


def main():
    initial = np.array(list(product([0, 1, -1], [0, 1, -1])))[1:]
    initial = normalize(initial)
    initial = np.hstack((initial, np.zeros((initial.shape[0], 1))))
    filename = "smooth_transition.html"
    figure = animate_init_(initial, filename)
    final = shear(initial)
    generation = 1
    figure = animate_next_(final, figure, filename, generation)
    complete = False
    while not complete:
        center = np.average(final, axis=0)
        final, complete = rotate_toward(center, np.array([1, 1, 1]), final)
        figure = animate_next_(final, figure, filename, generation)
        generation = generation + 1
    complete = False
    while not complete:
        center = np.average(final, axis=0)
        final, complete = rotate_toward(center, np.array([1, 0, 1]), final)
        figure = animate_next_(final, figure, filename, generation)
        generation = generation + 1
    complete = False
    while not complete:
        center = np.average(final, axis=0)
        final, complete = rotate_toward(center, np.array([0, 0, 1]), final)
        figure = animate_next_(final, figure, filename, generation)
        generation = generation + 1
    complete = False
    while not complete:
        center = np.average(final, axis=0)
        final, complete = rotate_toward(center, np.array([1, 0, 0]), final)
        figure = animate_next_(final, figure, filename, generation)
        generation = generation + 1


if __name__ == "__main__":
    main()
